import { Component, ViewChild, AfterContentInit } from "@angular/core";

import { Observer } from "jalhyd";

import { VarResults, GraphType } from "../../results/var-results";
import { GraphTypeSelectComponent } from "./graph-type.component";
import { ApplicationSetupService } from "../../services/app-setup/app-setup.service";

@Component({
    selector: "results-graph",
    templateUrl: "./results-graph.component.html",
    styleUrls: [
        "./results-graph.component.scss"
    ]
})
export class ResultsGraphComponent implements AfterContentInit, Observer {
    private _results: VarResults;

    @ViewChild(GraphTypeSelectComponent)
    private _graphTypeComponent: GraphTypeSelectComponent;

    /*
     * config du graphe
     */
    public graph_type: string;
    public graph_data = {};
    public graph_options = {
        responsive: true,
        maintainAspectRatio: true,
        animation: {
            duration: 0
        },
        legend: {
            display: false
        },
        title: {
            display: true,
            text: ""
        }
    };

    /** tracks the fullscreen state */
    public get isFullscreen() {
        return (document["fullscreenElement"] !== null);
    }

    public constructor(
        private appSetup: ApplicationSetupService
    ) {
        // limit display precision according to app preferences
        const nDigits = this.appSetup.displayDigits;
    }

    public set results(r: VarResults) {
        this._results = r;
        if (this._results && this._graphTypeComponent) {
            this._graphTypeComponent.selectedValue = r.graphType;
        }
    }

    public updateView() {
        switch (this._graphTypeComponent.selectedValue) {
            case GraphType.Histogram:
                this.graph_type = "bar";
                this.generateBarGraph();
                break;

            case GraphType.HistoLine:
                this.graph_type = "line";
                this.generateLineGraph();
                break;

            default:
                this.graph_type = "scatter";
                this.generateScatterGraph();
                break;
        }
    }

    public ngAfterContentInit() {
        this._graphTypeComponent.addObserver(this);
    }

    /**
     * génère les données d'un graphe de type "line"
     */
    private generateLineGraph() {
        const labs = [];
        const dat = [];
        let i = 0;
        for (const x of this._results.variatedParameter.valuesIterator) {
            labs.push(x);
            const y = this._results.yValues[i];
            dat.push(y);
            i++;
        }

        this.graph_options.title.text = this._results.graphTitle;

        this.graph_data = {
            labels: labs,
            datasets: [{
                label: "",
                data: dat
            }]
        };
    }

    /**
     * génère les données d'un graphe de type "bar"
     */
    private generateBarGraph() {
        const labs = [];
        const dat = [];
        let i = 0;
        for (const x of this._results.variatedParameter.valuesIterator) {
            labs.push(x);
            const y = this._results.yValues[i];
            dat.push(y);
            i++;
        }

        this.graph_options.title.text = this._results.graphTitle;
        const nDigits = this.appSetup.displayDigits;
        this.graph_options["scales"] = {
            xAxes: [{
                gridLines: {
                    offsetGridLines: true
                },
                ticks: {
                    precision: nDigits,
                    callback: function(value, index, values) {
                        return Number(value).toFixed(nDigits);
                    }
                }
            }]
        };
        this.graph_options["tooltips"] = {
            callbacks: {
                label: function(tooltipItem, data) {
                    return Number(tooltipItem.yLabel).toFixed(nDigits);
                }
            }
        };

        this.graph_data = {
            labels: labs,
            datasets: [{
                label: "",
                data: dat
            }]
        };
    }

    /**
     * génère les données d'un graphe de type "scatter"
     */
    private generateScatterGraph() {
        const dat = [];
        let i = 0;
        for (const x of this._results.variatedParameter.valuesIterator) {
            const y = this._results.yValues[i];
            dat.push({
                x: x,
                y: y
            });
            i++;
        }

        this.graph_options.title.text = this._results.graphTitle;
        const nDigits = this.appSetup.displayDigits;
        this.graph_options["scales"] = {
            xAxes: [{
                type: "linear",
                position: "bottom",
                ticks: {
                    precision: nDigits
                }
            }],
            yAxes: [{
                type: "linear",
                position: "left",
                ticks: {
                    precision: nDigits
                }
            }]
        };
        this.graph_options["tooltips"] = {
            callbacks: {
                label: function(tooltipItem, data) {
                    return "(" + Number(tooltipItem.xLabel).toFixed(nDigits) + ", " + Number(tooltipItem.yLabel).toFixed(nDigits) + ")";
                }
            }
        };

        this.graph_data = {
            datasets: [{
                label: "",
                data: dat,
                borderColor: "#808080", // couleur de la ligne
                backgroundColor: "rgba(0,0,0,0)",  // couleur de remplissage sous la courbe : transparent
                showLine: "true"
            }]
        };
    }

    public setFullscreen(element) {
        element.requestFullscreen();
    }

    public exitFullscreen() {
        document.exitFullscreen();
    }

    public exportAsImage(element: HTMLDivElement) {
        const canvas: HTMLCanvasElement = element.querySelector("canvas");
        canvas.toBlob((blob) => {
            saveAs(blob, "chart.png");
        }); // defaults to image/png
    }

    // interface Observer

    update(sender: any, data: any) {
        this.updateView();
    }
}
