import { CalculatorType, ComputeNodeType, ParamDefinition, LoiDebit, StructureType, Props, Observer, Nub, MethodeResolution } from "jalhyd";

import { FormulaireElement } from "./formulaire-element";
import { Field } from "./field";
import { CheckField } from "./check-field";
import { SelectField } from "./select-field";
import { NgParameter, ParamRadioConfig } from "./ngparam";
import { ServiceFactory } from "../services/service-factory";
import { ParamService } from "../services/param/param.service";
import { FormulaireDefinition } from "./definition/form-definition";
import { StringMap } from "../stringmap";
import { FormulaireNode } from "./formulaire-node";

export class FieldSet extends FormulaireElement implements Observer {
    /**
     * Nub associé
     */
    private _nub: Nub;

    /**
     * dictionnaire de traduction
     */
    private _localisation: StringMap;

    /**
     * fichier de configuration
     */
    private _jsonConfig: {};

    /**
     * propriétés déterminant l'état actuel du FieldSet (type de section, loi de débit, ...)
     */
    // private _props: Props;

    constructor(parent: FormulaireNode) {
        super(parent);
        // this._props = new Props();
    }

    public get nub(): Nub {
        return this._nub;
    }

    public setNub(sn: Nub, update: boolean = true) {
        this._nub = sn;
        this.setParentNubForAllFields();
        // this._props.setProps(sn.properties || new Props({}), this);
        if (update) {
            this.updateFields();
        }
    }

    private addField(f: Field) {
        if (! f) {
            throw new Error("FieldSet.addField() : argument incorrect (undefined)");
        }
        this.kids.push(f);
    }

    public get hasInputs(): boolean {
        for (const f of this.kids) {
            if (f instanceof NgParameter) {
                return true;
            }
        }
        return false;
    }

    public getInput(i: number): NgParameter {
        let n = 0;
        for (const f of this.kids) {
            if (f instanceof NgParameter) {
                if (n === i) {
                    return f;
                }
                n++;
            }
        }
    }

    public get isValid(): boolean {
        let res = true;
        for (const f of this.kids) {
            if (f instanceof Field) {
                if (f.isDisplayed) {
                    res = res && f.isValid;
                }
            }
        }
        return res;
    }

    private parse_check(json: {}): CheckField {
        const res: CheckField = new CheckField(this);
        res.parseConfig(json);
        return res;
    }

    private parse_select(json: {}): SelectField {
        const res: SelectField = new SelectField(this);
        res.parseConfig(json);
        res.addObserver(this);
        return res;
    }

    public get properties(): Props {
        return this.nub.properties;
        // return this._props;
    }

    private getPropValue(key: string): any {
        return this.properties.getPropValue(key);
    }

    public setPropValue(key: string, val: any): boolean {
        return this.properties.setPropValue(key, val, this);
    }

    private getNubParamFromSymbol(symbol: string): ParamDefinition {
        return this._nub.getParameter(symbol);
    }

    public get jsonConfig(): {} {
        return this._jsonConfig;
    }

    private setParentNubForAllFields() {
    }

    /**
     * crée un input
     * @param json definition de l'input, extrait du fichier de conf du module de calcul
     * @param node_type_filter filtre sur le type de noeud (input créé si undefined ou égal)
     * @param default_radio_config config du radio fixé/à varier/à calculer
     */
    private parse_input(json: {}, default_radio_config: string): NgParameter {
        // console.log("parsing input", json);
        const input_id: string = json["id"];
        const paramService: ParamService = ServiceFactory.instance.paramService;

        let res: NgParameter;

        switch (input_id) {
            case "Pr":
                res = paramService.createParameter(input_id, this);
                break;

            default:
                const nt: string = json["nodeType"];
                const nodeType: ComputeNodeType = nt ? ComputeNodeType[nt] : this.getPropValue("nodeType");

                if (nodeType === this.getPropValue("nodeType")) { // si le nodeType du paramètre est le même que celui du fieldset
                    let nubParam: ParamDefinition;
                    try {
                        nubParam = this.getNubParamFromSymbol(input_id);
                    } catch (e) {
                    }
                    if (nubParam) {
                        res = new NgParameter(nubParam, this);
                    }
                }
        }

        if (res) {
            res.parseConfig(json, { "radioConfig": default_radio_config });
            // set parent Nub on Parameter to ensure UID availability
            res.paramDefinition.parent = this._nub;
        }

        return res;
    }

    private parseFields() {
        const fields = this._jsonConfig["fields"];
        for (const field_index in fields) {
            const field = fields[field_index];

            if (field["type"] === "input") {
                const default_radio_config = this._jsonConfig["option"];
                const param = this.parse_input(field, default_radio_config);
                // tslint:disable-next-line:max-line-length
                if (param) { // potentiellement undefined car certaines définitions de FieldSet comportent des paramètres qui ne sont pas tous affichés en même temps (cf. ouvrages //)
                    this.addField(param);
                }
            } else if (field["type"] === "select") {
                const param = this.parse_select(field);
                this.addField(param);
            } else if (field["type"] === "check") {
                const param = this.parse_check(field);
                this.addField(param);
            }
        }
    }

    private clearFields() {
        for (const n of this.kids) {
            n.removeObserver(this);
        }

        this.clearKids();
    }

    public updateLocalisation(loc?: StringMap) {
        if (! loc) {
            loc = this._localisation;
        } else {
            this._localisation = loc;
        }

        if (loc) {
            super.updateLocalisation(loc);
        }
    }

    /**
     * Reflects all properties values in the interface, through the values of the <select> fields
     */
    private updateFields() {
        this.clearFields();
        this.parseFields();
        this.parseDependencies(this._jsonConfig);
        this.updateLocalisation();

        // MAJ des selects avec les valeurs actuelles des propriétés
        // spécifique à chaque modul de calcul, à revoir
        switch (this._confId) {

            case "fs_ouvrage":
                this.setSelectValueFromProperty("select_ouvrage", "structureType");

                const st: StructureType = this.getPropValue("structureType");
                switch (st) {
                    case StructureType.SeuilRectangulaire:
                        this.setSelectValueFromProperty("select_loidebit1", "loiDebit");
                        break;

                    case StructureType.VanneRectangulaire:
                        this.setSelectValueFromProperty("select_loidebit2", "loiDebit");
                        break;
                }
                break;

            case "fs_section":
                this.setSelectValueFromProperty("select_section", "nodeType");
                break;

            case "fs_param_calc":
                this.setSelectValueFromProperty("select_resolution", "methodeResolution");
                break;

            case "fs_target_data":
                this.setSelectValueFromProperty("select_target", "varCalc");
                break;

        }

        // fin MAJ selects
        this.applyDependencies();
    }

    /**
     * Reflects a property value in the interface, through the value of a <select> field, if this select exists
     */
    private setSelectValueFromProperty(selectId: string, propertyKey: string) {
        const selectField: SelectField = this.getFormulaireNodeById(selectId) as SelectField;
        if (selectField) {
            const propVal: any = this.getPropValue(propertyKey);
            const selectElement = selectField.getSelectedEntryFromValue(propVal); // @TODO juste setValue() ?
            try {
                selectField.setValue(selectElement);
            } catch (e) {
                console.error(`setSelectValueFromProperty: cannot set value ${propVal} on <select> ${selectId}`);
            }
        }
    }

    /**
     * Sets Nub default property from config file, unless this property is already set
     */
    private setPropertyValueFromConfig(json: {}, configKey: string, propertyKey: string, enumClass?) {
        const configValue: string = json[configKey];
        const currentValue = this.properties.getPropValue(propertyKey);
        if (configValue && ! currentValue) {
            let formalValue =  configValue;
            if (enumClass) {
                formalValue = enumClass[configValue];
            }
            this.setPropValue(propertyKey, formalValue);
        }
    }

    /**
     * Sets Nub default properties from config file, unless properties values are already set
     * (when deserialising an existing Nub)
     * @TODO default values should be held by model, not interface !
     */
    public parseConfig(json: {}, data?: {}) {
        this._jsonConfig = json;

        this._confId = json["id"];

        const parentForm = this.parentForm as FormulaireDefinition;
        const ct: string = json["calcType"];
        const currentCt = this.properties.getPropValue("calcType");
        const calc_type: CalculatorType = currentCt ? currentCt : (ct ? CalculatorType[ct] : parentForm.calculatorType);
        this.setPropValue("calcType", calc_type);

        const dnt: string = json["defaultNodeType"];
        const currentNt = this.properties.getPropValue("nodeType");
        const node_type: ComputeNodeType = currentNt ? currentNt : (dnt ? ComputeNodeType[dnt] : parentForm.nodeType);
        this.setPropValue("nodeType", node_type);

        this.setPropertyValueFromConfig(json, "defaultStructType", "structureType", StructureType);
        this.setPropertyValueFromConfig(json, "defaultLoiDebit", "loiDebit", LoiDebit);
        this.setPropertyValueFromConfig(json, "methodeResolution", "methodeResolution", MethodeResolution);
        this.setPropertyValueFromConfig(json, "varCalc", "varCalc");

        this.updateFields();
    }

    public parseDependencies(json: {}) {
        super.parseDependencies(json);

        for (const k1 in json) {
            if (k1 === "fields") {
                const fields = json[k1];
                for (const k2 in fields) {
                    const field = fields[k2];
                    switch (field["type"]) {
                        case "input":
                        case "select":
                        case "check":
                            for (const k of this.kids) {
                                if (k.id === field["id"]) {
                                    k.parseDependencies(field);
                                    break;
                                }
                            }
                            break;
                    }
                }
            }
        }
    }

    public getNodeParameter(symbol: string): NgParameter {
        for (const p of this.kids) {
            if (p instanceof NgParameter) {
                if (p.isDisplayed && p.symbol === symbol) {
                    return p;
                }
            }
        }
    }

    public getNodeParameterValue(symbol: string): number {
        const p = this.getNodeParameter(symbol);
        if (! p) {
            throw new Error(`FieldSet.getNodeParameterValue() : pas de paramètre ${symbol} trouvé`);
        }

        switch (p.radioState) {
            case ParamRadioConfig.FIX:
                return p.getValue();

            case ParamRadioConfig.VAR:
            case ParamRadioConfig.CAL:
                return undefined;
        }
    }

    /**
     * retourne la valeur actuellement sélectionnée d'un SelectField (qui doit être affiché)
     * @param selectFieldId id du SelectField
     * @returns valeur courante du select sans le préfixe
     */
    public getSelectedValue(selectFieldId: string): string {
        for (const p of this.kids) {
            if (p instanceof SelectField && p.isDisplayed && p.id === selectFieldId) {
                const value: string = p.getValue().value;
                return FormulaireElement.removePrefix(value, selectFieldId + "_");
            }
        }
    }

    // interface Observer

    update(sender: any, data: any) {
        if (data.action) {
            switch (data.action) {
                case "select":
                    const senderId: string = sender.id.replace(/\d+$/, "");
                    switch (senderId) {
                        case "select_section": // sections paramétrées
                            this.setPropValue("nodeType", data.value.value);
                            break;
                        case "select_ouvrage": // ouvrages parallèles
                            this.setPropValue("structureType", data.value.value);
                            break;
                        case "select_loidebit": // ouvrages parallèles et dérivés
                            this.setPropValue("loiDebit", data.value.value);
                            break;
                        case "select_resolution": // courbes de remous, méthode de résolution
                            this.setPropValue("methodeResolution", data.value.value);
                            break;
                        case "select_target": // courbes de remous, variable à calculer
                            this.setPropValue("varCalc", data.value.value);
                            break;
                    }
                    break;
            }
        }
    }

    /**
     * MAJ des liens entre paramètres lors de la désérialisation
     * @param json conf du fieldset issue du fichier
     * @param uidMap table de correspondance uid dans le fichier de conf <-> uid en mémoire
     */
    public updateParamsLinks(json: {}, uidMap: {}[]) {
        for (const ks in json) {
            switch (ks) {
                case "elements":
                    for (const e of json[ks]) {
                        if (Object.keys(e)[0] === "param") {
                            const prm = e["param"];
                            if (prm["values"]["mode"] === "LINK") {
                                // id du formulaire cible dans le fichier
                                const oldFormUid = +prm["values"]["form_uid"];

                                // correspondance avec l'objet mémoire
                                let newFormUid: string;
                                for (const m of uidMap) {
                                    if (m["type"] === "form" && m["old"] === oldFormUid) {
                                        newFormUid = m["new"];
                                        break;
                                    }
                                }

                                // formulaire dont le Nub est la cible du lien
                                const destForm: FormulaireDefinition
                                    = ServiceFactory.instance.formulaireService.getFormulaireFromId(newFormUid);

                                // paramètre source (celui qui est lié à une valeur)
                                const src: NgParameter = this.getFormulaireNodeById(prm["id"]) as NgParameter;

                                // création du lien
                                src.paramDefinition.defineReference(destForm.currentNub, prm["values"]["ref"]);
                            }
                            break;
                        }
                    }
            }
        }
    }
}
